--[[
File: Role.lua
Author: Zormus
--]]

local Utility = Library.LibZormus.Utility
local _, Roles = ...

Roles.Role = {}
local Role = Roles.Role
Role.__index = Role

local list
local id = {}
local function GetRoleName(slot)
	if not id[slot] then
		id[slot] = string.format("h%016X", slot - 1)
	end
	return list[id[slot]]
end

function Roles.UpdateList(roles)
	list = Inspect.Role.List()
	if not GetRoleName(1) then
		list[string.format("h%016X", 0)] = "Role 1"
	end
	local slot = #roles + 1
	while GetRoleName(slot) do
		roles[slot] = Role.Create(slot)
		slot = slot + 1
	end
end

function Role.Create(slot)
	local role = setmetatable({}, Role)
	role.slot = slot
	_roles[slot] = _roles[slot] or {}
	role.data = _roles[slot]
	local defaultData =
	{
		menu = true,
		type = "none",
		Left =
		{
			equip = nil,
			items = {}
		},
		Right =
		{
			equip = nil,
			items = {}
		},
		custom = {}
	}
	Utility.SetDefaultValues(role.data, defaultData)	
	role.macros = {}
	role:UpdateMacros()
	return role
end

function Role:GetMenu()
	return self.data.menu
end

function Role:SetMenu(menu)
	self.data.menu = menu
end

function Role:GetSlot()
	return self.slot
end

function Role:GetName()
	return GetRoleName(self.slot)
end

function Role:GetType()
	return self.data.type
end

function Role:SetType(type)
	self.data.type = type or "none"
end

function Role:GetEquip(button)
	return self.data[button].equip
end

function Role:SetEquip(button, set)
	self.data[button].equip = set
end

function Role:GetItems(button)
	return self.data[button].items
end

function Role:SetItems(button, items)
	self.data[button].items = items
end

function Role:GetCustom()
	return self.data.custom
end

function Role:SetCustom(custom)
	self.data.custom = custom
end

function Role:GetMacro(button)
	return self.macros[button] or self.macros["Left"]
end

function Role:UpdateMacros()
	for _, button in ipairs({"Left", "Right"}) do
		local macro = string.format("role %i", self:GetSlot())
		if self:GetEquip(button) then
			macro = string.format("%s\nloadequip %i\nwait 1", macro, self:GetEquip(button))
			for _, item in ipairs(self:GetItems(button)) do
				macro = string.format("%s\nequip %s", macro, item)
			end
		end
		for _, custom in ipairs(self:GetCustom()) do
			macro = string.format("%s\n%s", macro, custom)
		end
		self.macros[button] = macro
	end
end

Roles.Compare = {}

function Roles.Compare.Slot(role1, role2)
	return role1:GetSlot() < role2:GetSlot()
end

function Roles.Compare.Name(role1, role2)
	return string.lower(role1:GetName()) < string.lower(role2:GetName())
end

function Roles.Compare.Type(role1, role2)
	if role1:GetType() ~= role2:GetType() then
		local order = {tank = 1, heal = 2, dps = 3, support = 4, none = 5}
		return order[role1:GetType()] < order[role2:GetType()]
	else
		return Roles.Compare.Name(role1, role2)
	end
end
